// cmerg.C Version 7.41  Copyleft (c) 杉本典夫
// Time-stamp: <2023-12-15 09:42:31 tac>
//
// 欠損ケースを補完しながら、2つのdansのデータファイルをパラレルマージする。
// mergコマンドとの違いは2つのファイルに共通するキー項目を比べ、
// 欠損ケースを補完しながらマージする点。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法 :g++ -O -s -W cmerg cmerg.C -I/usr/local/include
//              -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "cmerg  Version 7.41          Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:cmerg  [<オプション> …] <データファイル-1> <データファイル-2>",
  "機能:欠損ケースを補完しながら2つのデータファイルをパラレルマージするんじゃ。",
  "     共通キー項目を比べ、欠損ケースを補完しながらマージするんじゃよ",
  "     ケースはキー項目順にソートされ、キー項目は最初だけ出力されるんじゃ。",
  "<オプション>",
  "  -cd[={-|+}]:ケースデータの出力指定じゃ。(-cd:-cd=+)",
  "    -:出力せず  +:出力(オプション省略時)",
  "  -f=<ファイル名>:指定したファイルからオプションを読み込むんじゃ。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するんじゃ。(-h:-h=man)",
  "  -id[={-|+}]:データ識別情報の出力指定じゃ。(-id:-id=+)",
  "    -:出力せず  +:出力(オプション省略時)",
  "  -key{1|2}=<第1キー項目番号>[,<第2キー項目番号>,…]",
  "    キー項目の番号を指定するんじゃ。このオプションは複数指定可能じゃよ。",
  "    key1:1番目のデータファイル  key2:2番目のデータファイル"
};

using lint = list<int>;                      // 整数リストの短名定義
using lint_citr = list<int>::const_iterator; // 反復子の短名定義

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  bool vcd;                     // ケースデータ出力フラッグ
  bool vid;                     // データ識別情報出力フラッグ
  lint vkeylst1;                // 第1データファイルのキー項目番号リスト
  lint vkeylst2;                // 第2データファイルのキー項目番号リスト
  vector<lint_citr> vkitr1vec;  // 第1キー項目番号反復子ベクトル
  vector<lint_citr> vkitr2vec;  // 第2キー項目番号反復子ベクトル
  int get_keylst(lint& keylst, const string& sstr);
  // 文字列からキー項目指定を読み込む
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  bool cd() const               // ケースデータ出力フラッグを返す
  { return vcd; }
  bool id() const               // データ識別情報力フラッグを返す
  { return vid; }
  const lint& keylst1() const // 第1データファイルのキー項目番号リストを返す
  { return vkeylst1; }
  const lint& keylst2() const // 第2データファイルのキー項目番号リストを返す
  { return vkeylst2; }
  lint_citr kitr1(size_t idx) const // 第1キー項目番号リスト反復子を返す
  { return (idx < vkitr1vec.size()) ? vkitr1vec[idx] : vkeylst1.end(); }
  lint_citr kitr2(size_t idx) const // 第2キー項目番号リスト反復子を返す
  { return (idx < vkitr2vec.size()) ? vkitr2vec[idx] : vkeylst2.end(); }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int nf, int ni);  // オプションのチェックとセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcd(true), vid(true), vkeylst1(), vkeylst2(), vkitr1vec(), vkitr2vec()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cd")           // ケースデータの出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vcd = true;
    else if (opt[1] == "-") vcd = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "id")      // データ識別情報の出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vid = true;
    else if (opt[1] == "-") vid = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "key1")    // 第1データファイルのキー項目の指定
  {
    if (get_keylst(vkeylst1, opt[1]) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "key2")    // 第2データファイルのキー項目の指定
  {
    if (get_keylst(vkeylst2, opt[1]) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::get_keylst(lint& keylst, const string& sstr)
// 文字列からキー項目指定を読み込む
{
  lstr kstrlst;
  if (strsplit(kstrlst, sstr, CHRdelm) < 1) return 0; // 有効な文字列なし
  size_t size0 = keylst.size();
  for (string kstr : kstrlst)
  {
    char* ptr = nullptr;
    int key = strtol(kstr.c_str(), &ptr, 0);
    if (key < 1) return -1;     // キー項目指定不適
    keylst.push_back(key);
  }
  return keylst.size() - size0; // 追加したキー項目指定数を返す
}

int option::check(int nf, int ni) // オプションのチェックとセット
{
  if (nf == 1)                  // 第1データファイルのキー項目チェック
  {
    if (vkeylst1.empty()) return disperr(vpname, "第1データファイルのキー項目が指定されとらん！");
    vkitr1vec.resize(vkeylst1.size());
    lint_citr kitr = vkeylst1.begin();
    int k = 0;
    while (kitr != vkeylst1.end())
    {
      if (*kitr > ni) return disperr(vpname, "第1データファイルのキー項目の指定が不適切！");
      vkitr1vec[k] = kitr;
      ++kitr;  ++k;
    }
  }
  else                          // 第2データファイルのキー項目チェック
  {
    if (vkeylst2.empty())       // キー項目指定がない場合
    { vkeylst2 = vkeylst1; }    // 第1キー項目をコピーする
    else if (vkeylst1.size() != vkeylst2.size())
    { return disperr(vpname, "2つのデータファイルのキー項目数が違っとるぞよ！"); }
    vkitr2vec.resize(vkeylst2.size());
    lint_citr kitr = vkeylst2.begin();
    int k = 0;
    while (kitr != vkeylst2.end())
    {
      if (*kitr > ni) return disperr(vpname, "第2データファイルのキー項目の指定が不適切！");
      vkitr2vec[k] = kitr;
      ++kitr;  ++k;
    }
  }
  return 0;                     // エラーなし
}

class key_less                  // キー項目の比較用関数オブジェクト
{
private:
  const lint *keylst;           // キー項目リストへのポインター
public:
  key_less(const lint *keylst0) : keylst(keylst0)
  // コンストラクタ
  {}
  bool operator()(const datac& cdata1, const datac& cdata2) const;
  // キー項目を比較する
};

bool key_less::operator()(const datac& cdata1, const datac& cdata2) const
// キー項目を比較する
{
  int comp = 0;
  lint_citr kitr = keylst->begin();
  while (comp == 0 && kitr != keylst->end())
  {
    int k = *kitr - 1;
    if ((comp = strcmp((cdata1.sdata(k)).c_str(), (cdata2.sdata(k)).c_str())) == 0) ++kitr;
  }
  return (comp < 0) ? true : false; // 小さいときだけtrueを返す
}

class dfile                     // データファイルクラス
{
private:
  string fname;                 // ファイル名
  idd vid;                      // データ識別情報
  ldata vdatalst;               // データリスト
  vector<ldata_itr> vitrvec;    // データリスト反復子ベクトル
  void set_itr();               // データリスト反復子配列をセットする
public:
  dfile() : fname(), vid(), vdatalst(), vitrvec()
  // デフォールトコンストラクタ
  {}
  dfile(const dfile& dfile0) : fname(dfile0.fname), vid(dfile0.vid), vdatalst(dfile0.vdatalst), vitrvec()
  // コピーコンストラクタ
  { set_itr(); }
  virtual ~dfile()              // デストラクタ
  {}
  const dfile& operator=(const dfile& dfile0); // コピー代入演算子
  int readf(const char* c_fname, option& opt, int nf);
  // データを読み込みメモリに保存する
  idd& id()                     // データ識別情報を返す
  { return vid; }
  ldata& datalst()              // データリストを返す
  { return vdatalst; }
  ldata_itr itr(size_t idx)     // データリスト反復子を返す
  { return (idx < vitrvec.size()) ? vitrvec[idx] : vdatalst.end(); }
};

const dfile& dfile::operator=(const dfile& dfile0) // コピー代入演算子
{
  if (&dfile0 == this) return *this; // 自分自身への代入を防ぐ
  fname = dfile0.fname;  vid = dfile0.vid;  vdatalst = dfile0.vdatalst;
  set_itr();
  return *this;
}

void dfile::set_itr()           // データリスト反復子配列をセットする
{
  vitrvec.resize(vdatalst.size());
  int i = 0;
  for (ldata_itr ditr = vdatalst.begin(); ditr != vdatalst.end(); ++ditr, ++i) vitrvec[i] = ditr;
}

int dfile::readf(const char* c_fname, option& opt, int nf)
// データを読み込みメモリに保存する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, vid, vdatalst);
  if (nc < 0) return nc;        // 実行時エラー
  if (vid.ni() < 1) return disperr(string("データがないでごんす!→") + fname);
  if (opt.check(nf, vid.ni()) != 0) return -2;
  if (nf == 1) vdatalst.sort(key_less(&(opt.keylst1())));
  else vdatalst.sort(key_less(&(opt.keylst2())));
  // ケースをキー項目順にソートする
  set_itr();
  return nc;                    // 読み込んだケース数を返す
}

class cmergclass                // cmergメインクラス
{
private:
  int check_key(size_t kno, const lint& keyls) const;
  // キー項目かどうかチェックする
  int comp_key(const option& opt, const datac* cdata1, const datac* cdata2) const;
  // キー項目を比較する
public:
  cmergclass()                  // コンストラクタ
  {}
  ~cmergclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, dfile& df1, dfile& df2, int nf) const;
  // データを読み込みケースをソートして保存する
  void put(const option& opt, dfile& df1, dfile& df2) const;
  // データを出力する
};

int cmergclass::check_key(size_t kno, const lint& keylst) const
// キー項目かどうかチェックする
{
  int kidx = -1, k = 0;
  lint_citr kitr = keylst.begin();
  while (kidx < 0 && kitr != keylst.end())
  {
    if ((int)kno == *kitr) kidx = k;
    else { ++kitr;  ++k; }
  }
  return kidx; // キー項目リスト反復子配列のインデックスを返す(キー項目以外:-1)
}

int cmergclass::comp_key(const option& opt, const datac* cdata1, const datac* cdata2) const
// キー項目を比較する
{
  int comp = 0;
  if (cdata1 == nullptr)        // 第1データが存在しない場合
  { comp = (cdata2 == nullptr) ? 0 : 1; }
  else if (cdata2 == nullptr)   // 第2データが存在しない場合
  { comp = -1; }
  else                          // 両方存在する場合
  {
    for (size_t k = 0; comp == 0 && k < (opt.keylst1()).size(); ++k)
    {
      int kidx1 = *(opt.kitr1(k)) - 1;
      int kidx2 = *(opt.kitr2(k)) - 1;
      comp = strcmp((cdata1->sdata(kidx1)).c_str(), (cdata2->sdata(kidx2)).c_str());
    }
  }
  return comp;    // comp<0:key1<key2,comp==0:key1==key2,comp>0:key1>key2
}

int cmergclass::get(const char* c_fname, option& opt, dfile& df1, dfile& df2, int nf) const
// データを読み込みケースをソートして保存する
{
  ++nf;
  int nc = 0;
  if (nf == 1) nc = df1.readf(c_fname, opt, nf);
  else nc = df2.readf(c_fname, opt, nf);
  if (nc < 0) throw runtime_error("実行時エラー！");
  return nf;                    // 読み込んだファイル数を返す
}

void cmergclass::put(const option& opt, dfile& df1, dfile& df2) const
// データを出力する
{
  if (opt.id())                 // データ識別情報出力
  {
    (df1.id()).put(cout, formtag);
    for (size_t j = 0; j < (df2.id()).ni(); ++j)
    {                           // 第2データファイルのキー項目名は出力しない
      if (check_key(j + 1, opt.keylst2()) < 0) (df2.id()).put_item(cout, j, formtag);
    }
  }
  if (opt.cd() == false) return;
  size_t i1 = 0, i2 = 0;
  datac *cdata1 = nullptr, *cdata2 = nullptr;
  while (i1 < (df1.id()).nc() || i2 < (df2.id()).nc())
  {
    if (i1 < (df1.id()).nc()) cdata1 = &(*(df1.itr(i1)));
    else cdata1 = nullptr;
    if (i2 < (df2.id()).nc()) cdata2 = &(*(df2.itr(i2)));
    else cdata2 = nullptr;
    int comp = comp_key(opt, cdata1, cdata2);
    bool putdelm = false;
    if (comp == 0)              // 両方のデータファイルにあるケース
    {
      for (size_t j = 0; j < (df1.id()).ni(); ++j)
      {
        if (putdelm) cout << CHRdelm;
        else putdelm = true;
        cout << cdata1->dstr(j);
      }
      for (size_t j = 0; j < (df2.id()).ni(); ++j)
      {                         // 第2データファイルのキー項目は出力しない
        if (check_key(j + 1, opt.keylst2()) < 0) cout << CHRdelm << cdata2->dstr(j);
      }
      ++i1;  ++i2;
    }
    else if (comp < 0)          // 第1データファイルだけにあるケース
    {
      for (size_t j = 0; j < (df1.id()).ni(); ++j)
      {
        if (putdelm) cout << CHRdelm;
        else putdelm = true;
        cout << cdata1->dstr(j);
      }
      for (size_t j = 0; j < (df2.id()).ni(); ++j)
      {
        if (check_key(j + 1, opt.keylst2()) < 0) cout << CHRdelm << BLANKs;
      }
      ++i1;
    }
    else                        // 第2データファイルだけにあるケース
    {
      for (size_t j = 0; j < (df1.id()).ni(); ++j)
      {
        if (putdelm) cout << CHRdelm;
        else putdelm = true;
        int kidx = check_key(j + 1, opt.keylst1());
        if (kidx < 0) cout << BLANKs; // キー項目ではない場合
        else cout << cdata2->dstr(*(opt.kitr2(kidx)) - 1); // 第2データファイルのキー項目データを出力する
      }
      for (size_t j = 0; j < (df2.id()).ni(); ++j)
      {                         // 第2データファイルのキー項目は出力しない
        if (check_key(j + 1, opt.keylst2()) < 0) cout << CHRdelm << cdata2->dstr(j);
      }
      ++i2;
    }
    if (putdelm) cout << endl;
  }
}

int main(int argc, char *argv[]) // メインルーチン
{
  cmergclass cmerg;             // メインクラス変数
  option opt(argv[0]);          // オプション
  dfile df1;                    // 第1データファイル
  dfile df2;                    // 第2データファイル
  try
  {
    opt.setdef();               // オプションの初期設定
    int nf = 0;                 // データファイル数
    for (int i = 1; i < argc; ++i) // 標準入力からは読み込まない
    {
      if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1); }
      else if (nf > 1)
      { throw runtime_error(string("データファイルが多すぎるんじゃ→") + string(argv[i])); }
      else
      { nf = cmerg.get(argv[i], opt, df1, df2, nf); }
    }
    if (nf == 2) cmerg.put(opt, df1, df2);
    else throw runtime_error("データファイルは2つ指定しないといかんのじゃ！");
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
