// lout.C Version 7.41  Copyleft (c) 杉本典夫
// Time-stamp: <2023-12-15 09:51:12 tac>
//
// dansデータをケースカード形式で出力する。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法 :g++ -O -s -W lout lout.C -I/usr/local/include
//              -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "lout  Version 7.41           Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:lout -f=<書式指定ファイル名> [<オプション> …] [<データファイル> …]",
  "機能:dansのデータをケースカード形式で出力します。",
  "<オプション>",
  "  -b=<ブランク文字列>:ヌル文字列を指定された文字列に変換して出力します。",
  "  -c=[<開始ケース番号>][-[<終了ケース番号>]]:出力ケース番号の指定。",
  "    -c=1-:全ケース出力(オプション省略時)",
  "  -f=<ファイル名>:指定したファイルからオプションを読み込みます。",
  "  -fmtf=<書式指定ファイル名>:指定したファイルから出力書式を読み込みます。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示します。(-h:-h=man)"
};

enum lfmttype                   // 書式の種類
{
  lfmtnon,                      // なし(ブランク)
  sdata,                        // 文字列
  sitem,                        // 文字項目データ
  ditem,                        // 整数項目データ
  fitem,                        // 実数項目データ
  newl,                         // 改行指定
  lfmtetc                       // その他
};

class lformat                   // ケースデータ形式の書式クラス
{
private:
  lfmttype vtype;               // 書式の種類
  int vidx;                     // 項目のインデックス
  string vstr;                  // 文字列
  int vn;                       // 全桁数
  int checkn(const string& str) const; // 桁指定をチェックする
public:
  lformat() : vtype(lfmtnon), vidx(-1), vstr(), vn(0) // デフォールトコンストラクタ
  {}
  lformat(lfmttype type0, int idx0, const string& str0);
  // コンストラクタ
  lformat(const lformat& lfmt) : vtype(lfmt.vtype), vidx(lfmt.vidx), vstr(lfmt.vstr), vn(lfmt.vn)
  // コピーコンストラクタ
  {}
  virtual ~lformat()            // デストラクタ
  {}
  const lformat& operator=(const lformat& lformat0); // コピー代入演算子
  lfmttype type() const         // 書式の種類を返す
  { return vtype; }
  lfmttype type(lfmttype type0) // 書式の種類を設定する
  { return (vtype = type0); }
  int idx() const               // 項目のインデックスを返す
  { return vidx; }
  int idx(int idx0)             // 項目のインデックスを設定する
  { return (vidx = idx0); }
  const string& str() const     // 文字列を返す
  { return vstr; }
  const string& str(const string& str0) // 文字列を設定する
  { return (vstr = str0); }
  int n() const                 // 出力桁数を返す
  { return vn; }
  int getlfmt(const string& str); // 文字列から書式を取り出す
};

lformat::lformat(lfmttype type0, int idx0, const string& str0) : vtype(type0), vidx(idx0), vstr(str0)
// コンストラクタ
{
  if (vtype == sdata)
  { vn = vstr.size(); }
  else
  {
    vn = checkn(vstr);
    if (vn == 0) vtype = lfmtnon; // 桁指定不適
  }
}

const lformat& lformat::operator=(const lformat& lformat0) // ゴー代入演算子
{
  if (&lformat0 == this) return *this; // 自分自身への代入を防ぐ
  vtype = lformat0.vtype;  vidx = lformat0.vidx;
  vstr = lformat0.vstr;  vn = lformat0.vn;
  return *this;
}

int lformat::checkn(const string& str) const // 桁指定をチェックする
{
  if (str.empty() || str[0] != '%') return 0; // 書式不適
  string tstr(str), substr;
  tstr.erase(0, 1);
  strsize_t idx = getsubstr(substr, tstr, '.');
  char* ptr = nullptr;
  int nt = strtol(substr.c_str(), &ptr, 0);
  if (nt == 0) return 0;        // 全桁指定不適
  if (idx < tstr.size())        // 小数部桁指定があるとき
  {
    tstr.erase(0, idx + 1);  ptr = 0;
    int f = strtol(tstr.c_str(), &ptr, 0);
    if (f < 0 || ptr == tstr.c_str()) nt = 0; // 小数部桁指定不適
  }
  return nt;                    // 全桁数を返す(正負あり)
}

int lformat::getlfmt(const string& str) // 文字列から書式を取り出す
{
  if (str.empty()) return -1;   // 文字列不適
  vtype = lfmtnon;  vidx = -1;  vstr.erase();  vn = 0;
  string tstr(str), substr;
  if (tstr[0] == CHRtags)       // タグ
  {
    tstr.erase(0, 1);
    strsize_t idx = getsubstr(substr, tstr, CHRetag);
    if (idx == string::npos || idx >= tstr.size() || substr.empty()) return -2;
    // タグ指定不適
    idx = getsubstr(tstr, substr, ' ');
    strtrm(tstr);
    if (tstr != TAGitem)        // 項目指定不適
    { return -3; }
    else
    {
      substr.erase(0, idx + 1);
      idx = getsubstr(tstr, substr, '=');
      char* ptr = nullptr;
      vidx = strtol(tstr.c_str(), &ptr, 0) - 1;
      if (idx == string::npos || idx >= substr.size() || vidx < 0 || ptr == tstr.c_str()) return -4; // 項目番号不適
      substr.erase(0, idx + 1);
      strtrm(substr);
      sstrtostr(substr);        // ""を削除する
      vn = checkn(substr);
      if (vn == 0) return -5;   // 桁数指定不適
      switch (substr[substr.size() - 1])
      {
      case 'd' : vtype = ditem;
        break;
      case 'f' : vtype = fitem;
        break;
      case 's' : vtype = sitem;
        break;
      default : vtype = lfmtetc;
      }
      vstr = substr;
    }
  }
  else                          // 文字列
  {
    sstrtostr(tstr);            // 特殊表現文字を変換し""を削除する
    vstr = tstr;
    vtype = sdata;  vn = vstr.size();
  }
  return 0;                     // 取り出し成功
}

using llformat = list<lformat>; // の別名定義書式リスト

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  string vblank;                // ブランク文字列
  pair<int, int> vc;            // 出力ケース番号
  llformat vlfmtlst;            // 書式リスト
  int fget_lfmt(const string& fname); // 書式をファイルから読み込む
  strsize_t get_lfmtstr(string& lfmtstr, const string& str) const;
  // 書式文字列をひとつ取り出す
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  const string& blank() const   // ブランク文字列を返す
  { return vblank; }
  int cs() const                // 出力開始ケース番号を返す
  { return vc.first; }
  int ce() const                // 出力終了ケース番号を返す
  { return vc.second; }
  const llformat& lfmtlst() const // 書式リストを返す
  { return vlfmtlst; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int ni) const;     // オプションのチェックとセット
};

option::option(const char* c_pname) : vpath(), vpname(), vblank(BLANKs), vc(1, INT_MAX), vlfmtlst()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "b")            // ブランク文字列指定
  {
    vblank = BLANKs;
    if (!(opt[1].empty()))
    {
      sstrtostr(opt[1]);
      vblank = opt[1];
    }
  }
  else if (opt[0] == "c")       // 出力ケース番号指定
  {
    if (getrange(vc.first, vc.second, opt[1]) < 1) err = badopt(vpname, optstr, USAGE, 1, INT_MAX);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "fmtf")    // 書式指定ファイル
  { err = fget_lfmt(opt[1]); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::fget_lfmt(const string& fname) // 書式をファイルから読み込む
{
  ifstream fin(fname.c_str());
  if (!fin) return disperr(vpname, "書式指定ファイルがありません→", fname);
  string sline;
  int err = 0;
  while (err == 0 && getsline(sline, fin) != EOF)
  {
    strchomp(sline);
    strsize_t idx = string::npos;
    string lfmtstr;
    lformat lfmt;
    while (!err && (idx = get_lfmtstr(lfmtstr, sline)) != string::npos)
    {
      if (lfmt.getlfmt(lfmtstr) != 0)
      { err = disperr(vpname, "書式指定が間違ってます→", lfmtstr); }
      else
      {
        vlfmtlst.push_back(lfmt);
        if (idx < sline.size()) sline.erase(0, idx);
        else break;             // 文字列終了
      }
    }
    if (!err)
    {
      lfmt.type(newl);          // 最後に改行指定を追加する
      vlfmtlst.push_back(lfmt);
    }
  }
  return err;
}

strsize_t option::get_lfmtstr(string& lfmtstr, const string& str) const
// 書式文字列をひとつ取り出す
{
  if (str.empty()) return string::npos;
  strsize_t idx = string::npos;
  if (str[0] == CHRtags)        // タグ
  {
    idx = getsubstr(lfmtstr, str, CHRtage);
    if (idx == string::npos || idx >= str.size()) return string::npos;
    lfmtstr += CHRtage;  ++idx;
  }
  else                          // 文字列
  { idx = getsubstr(lfmtstr, str, CHRtags); }
  return idx;                   // 次の書式文字列へのポインターを返す
}

int option::check(int ni) const // オプションのチェックとセット
{
  if (vlfmtlst.empty()) return disperr(vpname, "書式指定がないので処理できません!");
  for (auto lfmti : vlfmtlst)
  {
    if (lfmti.idx() >= ni) return disperr(vpname, "書式指定中の項目指定が不適です!");
  }
  return 0;                     // エラーなし
}

class loutclass                 // loutメインクラス
{
private:
  string fname;                 // ファイル名
  void put_lcase(ostream& fout, const option& opt, datac& cdata) const;
  // ケースデータをケースカード形式で出力する
public:
  loutclass() : fname()         // コンストラクタ
  {}
  ~loutclass()                  // デストラクタ
  {}
  int get(const char* c_fname, const option& opt, idd& id, ldata& datalst);
  // データを読み込みメモリに保存する
  void put(const option& opt, idd& id, ldata& datalst) const;
  // データを出力する
};

void loutclass::put_lcase(ostream& fout, const option& opt, datac& cdata) const
// ケースデータをケースカード形式で出力する
{
  for (llformat::const_iterator lfmtitr = (opt.lfmtlst()).begin(); lfmtitr != (opt.lfmtlst()).end(); ++ lfmtitr)
  {
    double fdata = BLANKf;
    string dstr;
    int npad = cdata.sdata(lfmtitr->idx()).size();
    char* c_tstr = new char[abs(lfmtitr->n()) + npad + 1];
    switch (lfmtitr->type())
    {
    case sdata : strcpy(c_tstr, (lfmtitr->str()).c_str()); // 文字列出力
      break;
    case sitem :                // 文字データ
      dstr = (cdata.sdata(lfmtitr->idx()) == BLANKs) ? opt.blank() : cdata.sdata(lfmtitr->idx());
      npad = dstr.size() - strclen(dstr, CHARwide);
      sprintf(c_tstr, "%*s", lfmtitr->n() + sgn((double)lfmtitr->n()) * npad, dstr.c_str());
      break;
    case ditem : fdata = cdata.fdata(lfmtitr->idx()); // 整数データ
      if (fdata == BLANKf) sprintf(c_tstr, "%*s", lfmtitr->n(), (opt.blank()).c_str());
      else sprintf(c_tstr, (lfmtitr->str()).c_str(), (int)fdata);
      break;
    case fitem : fdata = cdata.fdata(lfmtitr->idx()); // 実数データ
      if (fdata == BLANKf) sprintf(c_tstr, "%*s", lfmtitr->n(), (opt.blank()).c_str());
      else sprintf(c_tstr, (lfmtitr->str()).c_str(), fdata);
      break;
    case newl : c_tstr[0] = '\n';  c_tstr[1] = '\0'; // 改行
      break;
    default : c_tstr[0] = '\0';
    }
    fout << c_tstr;
    delete [] c_tstr;
  }
  fout << "\n" << endl;         // 最後に空行を2行出力する
}

int loutclass::get(const char* c_fname, const option& opt, idd& id, ldata& datalst)
// データを読み込みメモリに保存する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (opt.check(id.ni()) < 0) throw option_error("実行時オプションエラー！");
  return nc;
}

void loutclass::put(const option& opt, idd& id, ldata& datalst) const
// データを出力する
{
  int cno = 1;
  bool reverse = false;
  if (opt.cs() > opt.ce())      // 逆順出力
  { reverse = true;  cno = datalst.size();  datalst.reverse(); }
  for (ldata_itr dpos = datalst.begin(); dpos != datalst.end(); ++dpos)
  {
    if ((reverse == false && opt.cs() <= cno && cno <= opt.ce()) || (reverse && opt.cs() >= cno && cno >= opt.ce()))
    {
      put_lcase(cout, opt, *dpos);
      if (reverse) --cno;
      else ++cno;
    }
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  loutclass lout;               // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) lout.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { lout.get(argv[i], opt, id, datalst);  fok = false; }
    }
    lout.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生!");
    return 255;
  }
  return 0;
}
